@@ -36,6 +36,8 @@ import io.undertow.util.Methods;
 import io.undertow.util.Protocols;
 import io.undertow.util.URLUtils;
 import io.undertow.util.BadRequestException;
+import io.undertow.util.Cookies;
+
 import org.xnio.OptionMap;
 
 import static io.undertow.util.Headers.ACCEPT_CHARSET_STRING;
@@ -162,7 +164,7 @@ public abstract class HttpRequestParser {
 
     private final int maxParameters;
     private final int maxHeaders;
-    private final boolean slashDecodingFlag;
+    private final boolean allowEncodedSlash;
     private final boolean decode;
     private final String charset;
     private final int maxCachedHeaderSize;
@@ -208,7 +210,7 @@ public abstract class HttpRequestParser {
     public HttpRequestParser(OptionMap options) {
         maxParameters = options.get(UndertowOptions.MAX_PARAMETERS, UndertowOptions.DEFAULT_MAX_PARAMETERS);
         maxHeaders = options.get(UndertowOptions.MAX_HEADERS, UndertowOptions.DEFAULT_MAX_HEADERS);
-        slashDecodingFlag = URLUtils.getSlashDecodingFlag(options);
+        allowEncodedSlash = options.get(UndertowOptions.ALLOW_ENCODED_SLASH, false);
         decode = options.get(UndertowOptions.DECODE_URL, true);
         charset = options.get(UndertowOptions.URL_CHARSET, StandardCharsets.UTF_8.name());
         maxCachedHeaderSize = options.get(UndertowOptions.MAX_CACHED_HEADER_SIZE, UndertowOptions.DEFAULT_MAX_CACHED_HEADER_SIZE);
@@ -293,9 +295,13 @@ public abstract class HttpRequestParser {
                     handleHeaderValue(buffer, currentState, builder);
                 }
             }
+            sanitazeListTypeHeaders(builder);
+            return;
+        } else {
+            handleStateful(buffer, currentState, builder);
+            sanitazeListTypeHeaders(builder);
             return;
         }
-        handleStateful(buffer, currentState, builder);
     }
 
     private void handleStateful(ByteBuffer buffer, ParseState currentState, HttpServerExchange builder) throws BadRequestException {
@@ -355,6 +361,13 @@ public abstract class HttpRequestParser {
         }
     }
 
+    private void sanitazeListTypeHeaders(final HttpServerExchange builder) {
+        //NOTE: this should be used only for HTTP1.x
+        if(!Cookies.isCrumbsAssemplyDisabled()) {
+            Cookies.assembleCrumbs(builder.getRequestHeaders());
+        }
+        //TODO: list type headers?
+    }
 
     abstract void handleHttpVerb(ByteBuffer buffer, final ParseState currentState, final HttpServerExchange builder) throws BadRequestException;
 
@@ -372,10 +385,6 @@ public abstract class HttpRequestParser {
     private static final int IN_PATH = 4;
     private static final int HOST_DONE = 5;
 
-    private static final int PATH_SEGMENT_START = 0;
-    private static final int PATH_DOT_SEGMENT = 1;
-    private static final int PATH_NON_DOT_SEGMENT = 2;
-
     /**
      * Parses a path value
      *
@@ -391,8 +400,6 @@ public abstract class HttpRequestParser {
         int canonicalPathStart = state.pos;
         boolean urlDecodeRequired = state.urlDecodeRequired;
 
-        int pathSubState = 0;
-
         while (buffer.hasRemaining()) {
             char next = (char) (buffer.get() & 0xFF);
             if(!allowUnescapedCharactersInUrl && !ALLOWED_TARGET_CHARACTER[next]) {
@@ -414,16 +421,9 @@ public abstract class HttpRequestParser {
             } else if (next == ';') {
                 state.parseState = parseState;
                 state.urlDecodeRequired = urlDecodeRequired;
+                state.pos = canonicalPathStart;
                 // store at canonical path the partial path parsed up until here
                 state.canonicalPath.append(stringBuilder.substring(canonicalPathStart));
-                if (parseState == IN_PATH && pathSubState == PATH_DOT_SEGMENT) {
-                    // Inside a dot-segment (".", ".."), we don't want to allow removal of the ';' character from
-                    // the path. This is to avoid path traversal issues - "/..;" should not be treated as "/..".
-                    state.canonicalPath.append(";");
-                }
-                state.stringBuilder.append(";");
-                // set position to end of path (possibly start of parameter name)
-                state.pos = state.stringBuilder.length();
                 // handle the path parameters
                 handlePathParameters(buffer, state, exchange);
                 // if state is PATH, it means that handlePathParameters found a / after parsing path parameters
@@ -454,18 +454,6 @@ public abstract class HttpRequestParser {
                 } else if (next == '/' && parseState != HOST_DONE) {
                     parseState = IN_PATH;
                 }
-
-                // This is helper state that tracks if the parser is currently in a path dot-segment (".", "..") or not.
-                if (parseState == IN_PATH) {
-                    if (next == '/') {
-                        pathSubState = PATH_SEGMENT_START;
-                    } else if (next == '.' && (pathSubState == PATH_SEGMENT_START || pathSubState == PATH_DOT_SEGMENT)) {
-                        pathSubState = PATH_DOT_SEGMENT;
-                    } else {
-                        pathSubState = PATH_NON_DOT_SEGMENT;
-                    }
-                }
-
                 stringBuilder.append(next);
             }
 
@@ -481,15 +469,10 @@ public abstract class HttpRequestParser {
             exchange.setRelativePath("/");
             exchange.setRequestURI(path, true);
         } else if (parseState < HOST_DONE && state.canonicalPath.length() == 0) {
-            final String decodedRequestPath = decode(path, urlDecodeRequired, state, slashDecodingFlag, false);
-            exchange.setRequestPath(decodedRequestPath);
-            exchange.setRelativePath(decodedRequestPath);
-            if(urlDecodeRequired && allowUnescapedCharactersInUrl) {
-                final String uri = decode(path, urlDecodeRequired, state, slashDecodingFlag, false);
-                exchange.setRequestURI(uri);
-            } else {
-                exchange.setRequestURI(path);
-            }
+            String decodedPath = decode(path, urlDecodeRequired, state, allowEncodedSlash, false);
+            exchange.setRequestPath(decodedPath);
+            exchange.setRelativePath(decodedPath);
+            exchange.setRequestURI(path, false);
         } else {
             handleFullUrl(state, exchange, canonicalPathStart, urlDecodeRequired, path, parseState);
         }
@@ -509,15 +492,10 @@ public abstract class HttpRequestParser {
 
     private void handleFullUrl(ParseState state, HttpServerExchange exchange, int canonicalPathStart, boolean urlDecodeRequired, String path, int parseState) {
         state.canonicalPath.append(path.substring(canonicalPathStart));
-        final String requestPath = decode(state.canonicalPath.toString(), urlDecodeRequired, state, slashDecodingFlag, false);
-        exchange.setRequestPath(requestPath);
-        exchange.setRelativePath(requestPath);
-        if(urlDecodeRequired && allowUnescapedCharactersInUrl) {
-            final String uri = decode(path, urlDecodeRequired, state, slashDecodingFlag, false);
-            exchange.setRequestURI(uri, parseState == HOST_DONE);
-        } else {
-            exchange.setRequestURI(path, parseState == HOST_DONE);
-        }
+        String thePath = decode(state.canonicalPath.toString(), urlDecodeRequired, state, allowEncodedSlash, false);
+        exchange.setRequestPath(thePath);
+        exchange.setRelativePath(thePath);
+        exchange.setRequestURI(path, parseState == HOST_DONE);
     }
 
 
@@ -531,7 +509,7 @@ public abstract class HttpRequestParser {
      */
     @SuppressWarnings("unused")
     final void handleQueryParameters(ByteBuffer buffer, ParseState state, HttpServerExchange exchange) throws BadRequestException {
-        final StringBuilder stringBuilder = state.stringBuilder;
+        StringBuilder stringBuilder = state.stringBuilder;
         int queryParamPos = state.pos;
         int mapCount = state.mapCount;
         boolean urlDecodeRequired = state.urlDecodeRequired;
@@ -545,15 +523,12 @@ public abstract class HttpRequestParser {
         //we encounter an encoded character
 
         while (buffer.hasRemaining()) {
-            final char next = (char) (buffer.get() & 0xFF);
+            char next = (char) (buffer.get() & 0xFF);
             if(!allowUnescapedCharactersInUrl && !ALLOWED_TARGET_CHARACTER[next]) {
                 throw new BadRequestException(UndertowMessages.MESSAGES.invalidCharacterInRequestTarget(next));
             }
             if (next == ' ' || next == '\t') {
-                String queryString = stringBuilder.toString();
-                if(urlDecodeRequired && this.allowUnescapedCharactersInUrl) {
-                    queryString = decode(queryString, urlDecodeRequired, state, slashDecodingFlag, false);
-                }
+                final String queryString = stringBuilder.toString();
                 exchange.setQueryString(queryString);
                 if (nextQueryParam == null) {
                     if (queryParamPos != stringBuilder.length()) {
@@ -607,9 +582,9 @@ public abstract class HttpRequestParser {
         state.mapCount = mapCount;
     }
 
-    private String decode(final String value, boolean urlDecodeRequired, ParseState state, final boolean slashDecodingFlag, final boolean formEncoded) {
+    private String decode(final String value, boolean urlDecodeRequired, ParseState state, final boolean allowEncodedSlash, final boolean formEncoded) {
         if (urlDecodeRequired) {
-            return URLUtils.decode(value, charset, slashDecodingFlag, formEncoded, state.decodeBuffer);
+            return URLUtils.decode(value, charset, allowEncodedSlash, formEncoded, state.decodeBuffer);
         } else {
             return value;
         }
@@ -624,7 +599,8 @@ public abstract class HttpRequestParser {
         boolean urlDecodeRequired = state.urlDecodeRequired;
         String param = state.nextQueryParam;
         final StringBuilder stringBuilder = state.stringBuilder;
-        int pos = state.pos;
+        stringBuilder.append(";");
+        int pos = stringBuilder.length();
 
         //so this is a bit funky, because it not only deals with parsing, but
         //also deals with URL decoding the query parameters as well, while also
