@@ -34,17 +34,18 @@ import java.util.Map;
 import java.util.Set;
 import java.util.function.Supplier;
 
-import javax.servlet.ServletException;
-import javax.servlet.ServletOutputStream;
-import javax.servlet.SessionTrackingMode;
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
+import jakarta.servlet.ServletException;
+import jakarta.servlet.ServletOutputStream;
+import jakarta.servlet.SessionTrackingMode;
+import jakarta.servlet.http.Cookie;
+import jakarta.servlet.http.HttpServletResponse;
+import jakarta.servlet.http.HttpSession;
 
 import io.undertow.UndertowLogger;
 import io.undertow.server.HttpServerExchange;
 import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.servlet.UndertowServletMessages;
+import io.undertow.servlet.core.DefaultCharsetMapping;
 import io.undertow.servlet.handlers.ServletRequestContext;
 import io.undertow.util.CanonicalPathUtils;
 import io.undertow.util.DateUtils;
@@ -55,16 +56,12 @@ import io.undertow.util.HttpString;
 import io.undertow.util.Protocols;
 import io.undertow.util.RedirectBuilder;
 import io.undertow.util.StatusCodes;
-import org.xnio.Bits;
 
 import static io.undertow.util.URLUtils.isAbsoluteUrl;
 
 /**
- * Implementation of {@code HttpServletResponse}.
- *
  * @author Stuart Douglas
  * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
- * @author Flavia Rainone
  */
 public final class HttpServletResponseImpl implements HttpServletResponse {
 
@@ -77,22 +74,15 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     private PrintWriter writer;
     private Integer bufferSize;
     private long contentLength = -1;
-
-    private int flags = 0;
-    // response is inside include
-    private static final int INSIDE_INCLUDE_FLAG          = 1 << 0x00;
-    // response is done
-    private static final int RESPONSE_DONE_FLAG           = 1 << 0x01;
-    // ignored flush has been performed
-    private static final int IGNORED_FLUSH_PERFORMED_FLAG = 1 << 0x02;
-    // prevents anything to be added to response
-    private static final int TREAT_AS_COMMITTED_FLAG      = 1 << 0x03;
-    // indicates that we are closing the response and no further content should be written
-    private static final int CONTENT_FULLY_WRITTEN_FLAG   = 1 << 0x04;
-    //if a content type has been set either implicitly or implicitly
-    private static final int CHARSET_SET_FLAG             = 1 << 0x05;
-
+    private boolean insideInclude = false;
     private Locale locale;
+    private boolean responseDone = false;
+
+    private boolean ignoredFlushPerformed = false;
+
+    private boolean treatAsCommitted = false;
+
+    private boolean charsetSet = false; //if a content type has been set either implicitly or implicitly
     private String contentType;
     private String charset;
     private Supplier<Map<String, String>> trailerSupplier;
@@ -109,7 +99,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void addCookie(final Cookie newCookie) {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
+        if (insideInclude) {
             return;
         }
         final ServletCookieAdaptor servletCookieAdaptor = new ServletCookieAdaptor(newCookie);
@@ -124,19 +114,9 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         return exchange.getResponseHeaders().contains(name);
     }
 
-    @Override
-    public String encodeUrl(final String url) {
-        return encodeURL(url);
-    }
-
-    @Override
-    public String encodeRedirectUrl(final String url) {
-        return encodeRedirectURL(url);
-    }
-
     @Override
     public void sendError(final int sc, final String msg) throws IOException {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
+        if(insideInclude) {
             //not 100% sure this is the correct action
             return;
         }
@@ -155,7 +135,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         exchange.setStatusCode(sc);
         if(src.isRunningInsideHandler()) {
             //all we do is set the error on the context, we handle it when the request is returned
-            flags |= TREAT_AS_COMMITTED_FLAG;
+            treatAsCommitted = true;
             src.setError(sc, msg);
         } else {
             //if the src is null there is no outer handler, as we are in an asnc request
@@ -168,7 +148,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         responseState = ResponseState.NONE;
         resetBuffer();
         exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
-        flags &= ~TREAT_AS_COMMITTED_FLAG;
+        treatAsCommitted = false;
         final String location = servletContext.getDeployment().getErrorPages().getErrorLocation(sc);
         if (location != null) {
             RequestDispatcherImpl requestDispatcher = new RequestDispatcherImpl(location, servletContext);
@@ -219,14 +199,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
                 } else {
                     current = "";
                 }
-                String precanonLocation = location;
-                String query = "";
-                int firstQuestionMark = location.indexOf("?");
-                if (firstQuestionMark >= 0) {
-                    precanonLocation = location.substring(0, firstQuestionMark);
-                    query = location.substring(firstQuestionMark);
-                }
-                realPath = CanonicalPathUtils.canonicalize(servletContext.getContextPath() + current + precanonLocation) + query;
+                realPath = CanonicalPathUtils.canonicalize(servletContext.getContextPath() + current + location);
             }
             String loc = exchange.getRequestScheme() + "://" + exchange.getHostAndPort() + realPath;
             exchange.getResponseHeaders().put(Headers.LOCATION, loc);
@@ -257,7 +230,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG)) {
+        if (insideInclude || ignoredFlushPerformed) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE)) {
@@ -279,7 +252,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG)) {
+        if (insideInclude || ignoredFlushPerformed || treatAsCommitted) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE) && !exchange.getResponseHeaders().contains(Headers.CONTENT_TYPE)) {
@@ -301,7 +274,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setStatus(final int sc) {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
+        if (insideInclude || treatAsCommitted) {
             return;
         }
         if (responseStarted()) {
@@ -310,14 +283,6 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         exchange.setStatusCode(sc);
     }
 
-    @Override
-    public void setStatus(final int sc, final String sm) {
-        setStatus(sc);
-        if(!Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) && servletContext.getDeployment().getDeploymentInfo().isSendCustomReasonPhraseOnError()) {
-            exchange.setReasonPhrase(sm);
-        }
-    }
-
     @Override
     public int getStatus() {
         return exchange.getStatusCode();
@@ -367,7 +332,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public String getContentType() {
         if (contentType != null) {
-            if (Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
+            if (charsetSet) {
                 return contentType + ";charset=" + getCharacterEncoding();
             } else {
                 return contentType;
@@ -389,7 +354,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public PrintWriter getWriter() throws IOException {
         if (writer == null) {
-            if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
+            if (!charsetSet) {
                 //servet 5.5
                 setCharacterEncoding(getCharacterEncoding());
             }
@@ -416,14 +381,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setCharacterEncoding(final String charset) {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted() || writer != null || isCommitted()) {
+        if (insideInclude || responseStarted() || writer != null || isCommitted()) {
             return;
         }
-        if (charset != null) {
-            flags |= CHARSET_SET_FLAG;
-        } else {
-            flags &= ~CHARSET_SET_FLAG;
-        }
+        charsetSet = charset != null;
         this.charset = charset;
         if (contentType != null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, getContentType());
@@ -437,7 +398,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setContentLengthLong(final long len) {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
+        if (insideInclude || responseStarted()) {
             return;
         }
         if(len >= 0) {
@@ -449,24 +410,20 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     boolean isIgnoredFlushPerformed() {
-        return Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG);
+        return ignoredFlushPerformed;
     }
 
     void setIgnoredFlushPerformed(boolean ignoredFlushPerformed) {
-        if (ignoredFlushPerformed) {
-            flags |= IGNORED_FLUSH_PERFORMED_FLAG;
-        } else {
-            flags &= ~IGNORED_FLUSH_PERFORMED_FLAG;
-        }
+        this.ignoredFlushPerformed = ignoredFlushPerformed;
     }
 
     private boolean responseStarted() {
-        return exchange.isResponseStarted() || Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG);
+        return exchange.isResponseStarted() || ignoredFlushPerformed || treatAsCommitted;
     }
 
     @Override
     public void setContentType(final String type) {
-        if (type == null || Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
+        if (type == null || insideInclude || responseStarted()) {
             return;
         }
         ContentTypeInfo ct = servletContext.parseContentType(type);
@@ -474,10 +431,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         boolean useCharset = false;
         if(ct.getCharset() != null && writer == null && !isCommitted()) {
             charset = ct.getCharset();
-            flags |= CHARSET_SET_FLAG;
+            charsetSet = true;
             useCharset = true;
         }
-        if(useCharset || !Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
+        if(useCharset || !charsetSet) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader());
         } else if(ct.getCharset() == null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader() + "; charset=" + charset);
@@ -515,7 +472,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void closeStreamAndWriter() throws IOException {
-        if(Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG)) {
+        if(treatAsCommitted) {
             return;
         }
         if (writer != null) {
@@ -568,28 +525,34 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         responseState = ResponseState.NONE;
         exchange.getResponseHeaders().clear();
         exchange.setStatusCode(StatusCodes.OK);
-        flags &= ~(TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG);
+        treatAsCommitted = false;
     }
 
     @Override
     public void setLocale(final Locale loc) {
-        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
+        if (insideInclude || responseStarted()) {
             return;
         }
         this.locale = loc;
         exchange.getResponseHeaders().put(Headers.CONTENT_LANGUAGE, loc.getLanguage() + "-" + loc.getCountry());
-        if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG) && writer == null) {
+        if (!charsetSet && writer == null) {
             final Map<String, String> localeCharsetMapping = servletContext.getDeployment().getDeploymentInfo().getLocaleCharsetMapping();
-            // Match full language_country_variant first, then language_country,
-            // then language only
-            String charset = localeCharsetMapping.get(locale.toString());
-            if (charset == null) {
-                charset = localeCharsetMapping.get(locale.getLanguage() + "_"
-                        + locale.getCountry());
+            // first try DD provided mappings
+            String charset = null;
+            if (!localeCharsetMapping.isEmpty()) {
+                charset = localeCharsetMapping.get(locale.toString());
                 if (charset == null) {
-                    charset = localeCharsetMapping.get(locale.getLanguage());
+                    charset = localeCharsetMapping.get(locale.getLanguage() + "_"
+                            + locale.getCountry());
+                    if (charset == null) {
+                        charset = localeCharsetMapping.get(locale.getLanguage());
+                    }
                 }
             }
+            // if DD provided mapping failed then use default
+            if (charset == null) {
+                charset = DefaultCharsetMapping.INSTANCE.getCharset(loc);
+            }
             if (charset != null) {
                 this.charset = charset;
                 if (contentType != null) {
@@ -609,10 +572,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void responseDone() {
-        if (Bits.anyAreSet(flags, RESPONSE_DONE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
+        if (responseDone || treatAsCommitted) {
             return;
         }
-        flags |= RESPONSE_DONE_FLAG;
+        responseDone = true;
         try {
             closeStreamAndWriter();
         } catch (IOException e) {
@@ -623,15 +586,11 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isInsideInclude() {
-        return Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG);
+        return insideInclude;
     }
 
     public void setInsideInclude(final boolean insideInclude) {
-        if (insideInclude) {
-            this.flags |= INSIDE_INCLUDE_FLAG;
-        } else {
-            this.flags &= ~INSIDE_INCLUDE_FLAG;
-        }
+        this.insideInclude = insideInclude;
     }
 
     public void setServletContext(final ServletContextImpl servletContext) {
@@ -820,7 +779,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isTreatAsCommitted() {
-        return Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG);
+        return treatAsCommitted;
     }
 
     @Override
@@ -851,11 +810,4 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         return trailerSupplier;
     }
 
-    /**
-     * Marks this response as closed for writing extra bytes, including the addition of headers.
-     */
-    void setContentFullyWritten() {
-        this.flags |= CONTENT_FULLY_WRITTEN_FLAG;
-    }
-
 }
